import {world, MinecraftDimensionTypes, BlockPermutation} from "@minecraft/server";
import * as debug from "debug.js"

// Tool code
export const clamp = (num, min, max) => Math.min(Math.max(num, min), max);
export const revolve = (num, limit) => Math.abs(num) > Math.abs(limit) ? (num > 0 ? limit - (num % limit) : limit*-1 - (num % limit*-1)) : num;
export const distance2D = (locA, locB) => ((((locB.x - locA.x) ** 2) + ((locB.y - locA.y) ** 2)) ** 1/2);
export const distance2Dz = (locA, locB) => ((((locB.x - locA.x) ** 2) + ((locB.z - locA.z) ** 2)) ** 1/2);
export const distance3D = (locA, locB) => ((((locB.x - locA.x) ** 2) + ((locB.y - locA.y) ** 2) + ((locB.z - locA.z) ** 2)) ** 1/2);
export const camel_to_snake_case = str => str.replace(/[A-Z]/g, letter => `_${letter.toLowerCase()}`);
export const min_bound = (num, bound) => (num > 0) ? Math.max(num, Math.abs(bound)) : Math.min(num, Math.abs(bound) * -1);
export const view_rot_to_dir = (viewRot) => ({x: -1 * Math.sin(viewRot.y * (Math.PI) / 180), y: -1 * viewRot.x / 90, z: Math.cos(viewRot.y * (Math.PI) / 180)});
export function parse_location(input) {
	// Fail if loc is empty
	if (input.length < 1)
		throw new Error("Error: sf_smf:guide_move requires a location of the format x,y,z");
	
	const coordinates = input.split(",");
	
	// Fail if we didn't get 3 items
	if (coordinates.length != 3)
		throw new Error("Error: sf_smf:guide_move requires a location of the format x,y,z");
	
	// Create return location
	const loc = {x: parseFloat(coordinates[0]), y: parseFloat(coordinates[1]), z: parseFloat(coordinates[2])};
	
	// Fail if a value isn't a number
	if (isNaN(loc.x))
		throw new Error(`Error: Provided x coordinate is invalid: ${coordinates[0]}`);
	if (isNaN(loc.y))
		throw new Error(`Error: Provided y coordinate is invalid: ${coordinates[1]}`);
	if (isNaN(loc.z))
		throw new Error(`Error: Provided y coordinate is invalid: ${coordinates[2]}`);
	
	// Return location
	return loc;
}
export const analyze = (obj) => {
	// Utility for analyzing objects.
	let properties = new Set()
	Object.getOwnPropertyNames(obj).map(item => properties.add(item))
	
	let currentObj = obj
	while ((currentObj = Object.getPrototypeOf(currentObj))) 
	{
		Object.getOwnPropertyNames(currentObj).map(item => properties.add(item))
	} 
	return [...properties.keys()]
}
export function get_all_subclasses(baseClass) {
	var allVars = Object.keys(world);
	var classes = allVars.filter(function (key) {
		try {
			var obj = world[key];
			return obj.prototype instanceof baseClass;
		}
		catch (e) 
		{
			console.warn(`Get classes error: ${err}`)
			return false;
		}
	});
	return classes;
}
export function get_entities(dimension, options) {
	if (options == null)
		options = {};
	
	let list = [];
	
	try {
		if (dimension == "all")
		{
			for (let dimension in MinecraftDimensionTypes) {
				list = [ ...list, ...world.getDimension(camel_to_snake_case(dimension)).getEntities(options)]
			};
		}
		else
			list = dimension.getEntities(options);
	}
	catch (err)
	{
		console.warn(`Get entities error: ${err}`)
	}
	return list
}
export function list_entity_components() {
	world.sendMessage(JSON.stringify(world.getDimension("overworld").getEntities().reduce((o, e) => {
		o[e.typeId.replace('minecraft:', '')] = e.getComponents().map(c => c?.typeId.substring(10))
		return o
	}, {}), null, 3));
}
export function direction_invert(dir) {
	switch (dir) {
		case "North":
			return "South";
		case "South":
			return "North";
		case "East":
			return "West";
		case "West":
			return "East";
		case "Up":
			return "Down";
		case "Down":
			return "Up";
		default:
			return "North";
	}
}
export function direction_detect(entity) {
	const direction = entity.getRotation().y;
	
	if ((direction >= -45) && (direction < 45))
		return "South";
	else if ((direction >= 45) && (direction < 135))
		return "West";
	else if ((direction >= 135) || (direction < -135))
		return "North";
	else if ((direction >= -135) && (direction < -45))
		return "East";
}

export function at_location(item, location) {
	return (item.location.x == location.x) && (item.location.y == location.y) && (item.location.z == location.z);
}

export function update_block_state(block, states) {
	const blockStates = block?.permutation.getAllStates();
	for (const state of Object.keys(states)) {
		if (blockStates[state] != null)
			blockStates[state] = states[state];
		else
			debug.warn(`Block ${block.typeId} doesn't support state "${state}". Ignoring...`);
	}
	block.setPermutation(BlockPermutation.resolve(block.typeId, blockStates));
}

export function count_item(player, type) {
	const container = player.getComponent("inventory").container;
	const size = container.size;
	let total = 0;
	for (let i = 0; i < size; i++) {
		const itemInSlot = container.getItem(i);
		if (itemInSlot?.typeId != type) continue; // If not correct item, skip
		else total += itemInSlot?.amount; // If correct item, increment total by num of items.
	}
	return total;
}
export function decrement_item(player, type, count) {
	const container = player.getComponent("inventory").container;
	const size = container.size;
	let total = count;
	for (let i = 0; i < size; i++) {
		if (total == 0)
			break;
		
		const itemInSlot = container.getItem(i);
		if (itemInSlot?.typeId != type) continue; // If not correct item, skip
		else {
			// If correct item, decrement total accordingly.
			const amount = itemInSlot?.amount;
			if (total >= amount) {
				container.setItem(i);
				total -= amount;
			}
			else {
				itemInSlot.amount -= total;
				container.setItem(i, itemInSlot);
				total = 0;
			}
		}
	}
}

export function string_adjust(str, index, item) {
	// Replaces character in str at index with item
    return str.substring(0, index) + item + str.substring(index + item.length);
}

export function give_item(player, itemStack) {
	const container = player.getComponent("inventory").container;
	
	system.runTimeout(() => {
		if (container.emptySlotsCount >= 1) {
			container.addItem(itemStack);
		}
		else {
			player.dimension.spawnItem(itemStack, player.location);
		}
	}, 2);
}

export function unequip_item(player, slot) {
	const equipment = player.getComponent("equippable");
	const item = equipment.getEquipment(slot);
	if (item != null) {
		give_item(player, item);
		equipment.setEquipment(slot);
	}
}

export function condensed_bool_read(variable, index) {
	return (Math.floor((variable/Math.pow(2,index)) % 2));
}

export function condensed_bool_write(variable, index, value) {
	if (value == 1)
		return (!condensed_bool_read(variable, index) ? variable + Math.pow(2, index) : variable);
	else if (value == 0)
		return (condensed_bool_read(variable, index) ? variable - Math.pow(2, index) : variable);
	else
		throw new Error(`Condensed var error: Cannot set value at index ${index} to ${value}. Unsafe action blocked.`);
}

export function is_moving(entity) {
	const velocity = entity.getVelocity();
	return (velocity.x != 0) || (velocity.y != 0) || (velocity.z != 0)
}

export function translate_text(text) {
	//const signPerm = BlockPermutation.resolve('minecraft:standing_sign', { ground_sign_direction: 8 });
    const signComponent = new BlockSignComponent();
    signComponent.setText({ translate: text });
	return signComponent.getText();
}

export function manual_parse_name(entity) {
	return manual_format_name_by_id(entity.typeId);
}

export function manual_format_name_by_id(text) {
	return text.substring(text.indexOf(":") + 1,text.length).replaceAll("_", " ").replace(/\w\S*/g, function (txt) {return txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase();})
}